"""Test keyboard invoked execution"""


from .utils import EDITOR_PAGE, validate_dualmode_state


INITIAL_CELLS = ['', 'print("a")', 'print("b")', 'print("c")']


def test_dualmode_execute(prefill_notebook):
    notebook_frontend = prefill_notebook(INITIAL_CELLS)
    for i in range(1, 4):
        notebook_frontend.execute_cell(i)

    # shift-enter
    # ...........
    # last cell in notebook
    base_index = 3
    notebook_frontend.focus_cell(base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE, ['Shift'])  # creates one cell
    validate_dualmode_state(notebook_frontend, 'edit', base_index + 1)
    # ...............................................
    # Not last cell in notebook & starts in edit mode
    notebook_frontend.focus_cell(base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE)  # Enter edit mode
    validate_dualmode_state(notebook_frontend, 'edit', base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE, ['Shift'])  # creates one cell
    validate_dualmode_state(notebook_frontend, 'command', base_index + 1)
    # ......................
    # Starts in command mode
    notebook_frontend.press('k', EDITOR_PAGE)
    validate_dualmode_state(notebook_frontend, 'command', base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE, ['Shift'])  # creates one cell
    validate_dualmode_state(notebook_frontend, 'command', base_index + 1)

    # Ctrl-enter
    # ..........
    # Last cell in notebook
    base_index += 1
    notebook_frontend.press("Enter", EDITOR_PAGE, [notebook_frontend.get_platform_modifier_key()])
    validate_dualmode_state(notebook_frontend, 'command', base_index)
    # ...............................................
    # Not last cell in notebook & stats in edit mode
    notebook_frontend.focus_cell(base_index - 1)
    notebook_frontend.press("Enter", EDITOR_PAGE)  # Enter edit mode
    validate_dualmode_state(notebook_frontend, 'edit', base_index - 1)
    notebook_frontend.press("Enter", EDITOR_PAGE, [notebook_frontend.get_platform_modifier_key()])
    # ...............................................
    # Starts in command mode
    notebook_frontend.press('j', EDITOR_PAGE)
    validate_dualmode_state(notebook_frontend, 'command', base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE, [notebook_frontend.get_platform_modifier_key()])
    validate_dualmode_state(notebook_frontend, 'command', base_index)

    # Alt-enter
    # ...............................................
    # Last cell in notebook
    notebook_frontend.press("Enter", EDITOR_PAGE, ['Alt'])
    validate_dualmode_state(notebook_frontend, 'edit', base_index + 1)
    # Not last cell in notebook &starts in edit mode
    notebook_frontend.focus_cell(base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE)  # Enter edit mode
    validate_dualmode_state(notebook_frontend, 'edit', base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE, ['Alt'])
    validate_dualmode_state(notebook_frontend, 'edit', base_index + 1)
    # starts in command mode
    notebook_frontend.press("Escape", EDITOR_PAGE)
    notebook_frontend.press('k', EDITOR_PAGE)
    validate_dualmode_state(notebook_frontend, 'command', base_index)
    notebook_frontend.press("Enter", EDITOR_PAGE, ['Alt'])
    validate_dualmode_state(notebook_frontend, 'edit', base_index + 1)

    # Notebook will now have 8 cells, the index of the last cell will be 7
    assert len(notebook_frontend.cells) == 8  # Cells were added
    notebook_frontend.focus_cell(7)
    validate_dualmode_state(notebook_frontend, 'command', 7)



"""Tests clipboard by copying, cutting and pasting multiple cells"""


from .utils import TREE_PAGE, EDITOR_PAGE


# Optionally perfom this test with Ctrl+c and Ctrl+v
def test_clipboard_multiselect(prefill_notebook):
    notebook = prefill_notebook(['', '1', '2', '3', '4', '5a', '6b', '7c', '8d'])

    assert notebook.get_cells_contents() == ['', '1', '2', '3', '4', '5a', '6b', '7c', '8d']

    # Copy the first 3 cells
    # Paste the values copied from the first three cells into the last 3 cells 

    # Selecting the fist 3 cells
    notebook.select_cell_range(1, 3)

    # Copy those selected cells
    notebook.try_click_selector('#editlink', page=EDITOR_PAGE)
    notebook.try_click_selector('//*[@id="copy_cell"]/a/span[1]', page=EDITOR_PAGE)

    # Select the last 3 cells
    notebook.select_cell_range(6, 8)

    # Paste the cells in clipboard onto selected cells
    notebook.try_click_selector('#editlink', page=EDITOR_PAGE)
    notebook.try_click_selector('//*[@id="paste_cell_replace"]/a', page=EDITOR_PAGE)

    assert notebook.get_cells_contents() == ['', '1', '2', '3', '4', '5a', '1', '2', '3']

    # Select the last four cells, cut them and paste them below the first cell

    # Select the last 4 cells
    notebook.select_cell_range(5, 8)

    # Click Edit button and the select cut button
    notebook.try_click_selector('#editlink', page=EDITOR_PAGE)
    notebook.try_click_selector('//*[@id="cut_cell"]/a', page=EDITOR_PAGE)

    # Select the first cell
    notebook.select_cell_range(0, 0)

    # Paste the cells in our clipboard below this first cell we are focused at
    notebook.try_click_selector('#editlink', page=EDITOR_PAGE)
    notebook.try_click_selector('//*[@id="paste_cell_below"]/a/span[1]', page=EDITOR_PAGE)

    assert notebook.get_cells_contents() == ['', '5a', '1', '2', '3', '1', '2', '3', '4']



from .utils import EDITOR_PAGE


def undelete(nb):
    nb.evaluate('() => Jupyter.notebook.undelete_cell();', page=EDITOR_PAGE)

INITIAL_CELLS = ['print("a")', 'print("b")', 'print("c")', 'print("d")']

def test_undelete_cells(prefill_notebook):
    notebook = prefill_notebook(INITIAL_CELLS)
    a, b, c, d = INITIAL_CELLS

    # Verify initial state
    assert notebook.get_cells_contents() == [a, b, c, d]

    # Delete cells [1, 2]
    notebook.focus_cell(1)
    notebook.press('j', EDITOR_PAGE, ['Shift'])
    notebook.press('d+d', EDITOR_PAGE)
    assert notebook.get_cells_contents() == [a, d]

    # Delete new cell 1 (which contains d)
    notebook.focus_cell(1)
    notebook.press('d+d', EDITOR_PAGE)
    assert notebook.get_cells_contents() == [a]

    # Undelete d
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, d]

    # Undelete b, c
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, b, c, d]

    # Nothing more to undelete
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, b, c, d]

    # Delete first two cells and restore
    notebook.focus_cell(0)
    notebook.press('j', EDITOR_PAGE, ['Shift'])
    notebook.press('d+d', EDITOR_PAGE)
    assert notebook.get_cells_contents() == [c, d]
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, b, c, d]

    # Delete last two cells and restore
    notebook.focus_cell(-1)
    notebook.press('k', EDITOR_PAGE, ['Shift'])
    notebook.press('d+d', EDITOR_PAGE)
    assert notebook.get_cells_contents() == [a, b]
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, b, c, d]

    # Merge cells [1, 2], restore the deleted one
    bc = b + "\n\n" + c
    notebook.focus_cell(1)
    notebook.press('j', EDITOR_PAGE, ['Shift'])
    notebook.press('m', EDITOR_PAGE, ['Shift'])
    assert notebook.get_cells_contents() == [a, bc, d]
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, bc, c, d]

    # Merge cells [2, 3], restore the deleted one
    cd = c + "\n\n" + d
    notebook.focus_cell(-1)
    notebook.press('k', EDITOR_PAGE, ['Shift'])
    notebook.press('m', EDITOR_PAGE, ['Shift'])
    assert notebook.get_cells_contents() == [a, bc, cd]
    undelete(notebook)
    assert notebook.get_cells_contents() == [a, bc, cd, d]

    # Reset contents to [a, b, c, d] --------------------------------------
    notebook.edit_cell(index=1, content=b)
    notebook.edit_cell(index=2, content=c)
    assert notebook.get_cells_contents() == [a, b, c, d]

    # Merge cell below, restore the deleted one
    ab = a + "\n\n" + b
    notebook.focus_cell(0)
    notebook.evaluate("() => Jupyter.notebook.merge_cell_below();", page=EDITOR_PAGE)
    assert notebook.get_cells_contents() == [ab, c, d]
    undelete(notebook)
    assert notebook.get_cells_contents() == [ab, b, c, d]

    # Merge cell above, restore the deleted one
    cd = c + "\n\n" + d
    notebook.focus_cell(-1)
    notebook.evaluate("() => Jupyter.notebook.merge_cell_above();", page=EDITOR_PAGE)
    assert notebook.get_cells_contents() == [ab, b, cd]
    undelete(notebook)
    assert notebook.get_cells_contents() == [ab, b, c, cd]



"""Test navigation to directory links"""


import os

from .utils import  TREE_PAGE
from jupyter_server.utils import url_path_join
pjoin = os.path.join


def url_in_tree(nb, url=None):
    if url is None:
        url = nb.get_page_url(page=TREE_PAGE)

    tree_url = url_path_join(nb.get_server_info(), 'tree')
    return True if tree_url in url else False


def get_list_items(nb):
    """
    Gets list items from a directory listing page
    """

    nb.wait_for_selector('#notebook_list .item_link', page=TREE_PAGE)
    notebook_list = nb.locate('#notebook_list', page=TREE_PAGE)
    link_items = notebook_list.locate_all('.item_link')

    return [{
        'link': a.get_attribute('href'),
        'label': a.get_inner_text(),
        'element': a,
    } for a in link_items if a.get_inner_text() != '..']


def test_navigation(notebook_frontend):
    print('[Test] [test_dashboard_nav] Start!')

    print('[Test] Obtain list of elements')
    link_elements = get_list_items(notebook_frontend)

    # Recursively traverse and check folder in the Jupyter root dir
    def check_links(nb, list_of_link_elements):
        print('[Test] Check links')
        if len(list_of_link_elements) < 1:
            return

        for item in list_of_link_elements:
            print(f'[Test]   Check "{item["label"]}"')
            if 'Untitled.ipynb' in item["label"]:
                # Skip notebook files in the temp dir
                continue

            item["element"].click()

            notebook_frontend.wait_for_condition(
                lambda: url_in_tree(notebook_frontend),
                timeout=600,
                period=5
            )
            notebook_frontend.wait_for_condition(
                lambda: item["link"] in nb.get_page_url(page=TREE_PAGE),
                timeout=600,
                period=5
            )

            new_links = get_list_items(nb)
            if len(new_links) > 0:
                check_links(nb, new_links)

            nb.go_back(page=TREE_PAGE)

        return

    check_links(notebook_frontend, link_elements)
    print('[Test] [test_dashboard_nav] Finished!')
